package massim.agent.framework;

import java.util.HashMap;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Structure holding the perceived fragment of the environment as the simulation server
 * passes it to agents in each simulation step. 
 */
public class Perception {

    /**
     * @return the sightRange
     */
    public static int getSightRange() {
        return sightRange;
    }
	
	/** 
	 * Agent's X position in the grid.
	 */
	public int posX;
	
	/**
	 * Agent's Y position in the grid.
	 */
	public int posY;
	
	/**
	 * The simulation step number.
	 */
	public int step;
	
	/**
	 * The simulation step timestamp, as received from the MASSim server.
	 * 
	 * For closer explanation (if you feel you need it), check the MASSim protocol specification.
	 */
	public long timestamp;
	
	/**
	 * The sight range of an agent.
	 * 
	 * The number determines the number of cells the agent does see in a straight line 
	 * in X, resp. Y, direction. In fact, the size of the square of cells the agent perceives
	 * in every step equals to 2*sightRange+1. That is, the agent can see a square of cells, 
	 * center of which is the agent's own position and to each side from that position it is 
	 * sightRange cells large. 
	 * 
	 * The number is set at the beginning of the simulation.
	 */
	private static int sightRange = 0;
	
	/**
	 * Horizontal size of the environment grid.
	 * 
	 * The number is set at the beginning of the simulation.
	 */
	public static int gridSizeX = 0;
	
	/**
	 * Vertical size of the environment grid.
	 * 
	 * The number is set at the beginning of the simulation.
	 */
	public static int gridSizeY = 0;
	
	/**
	 * The overall number of steps the simulation should take. 
	 * 
	 * The number is set at the beginning of the simulation.
	 */
	public static int simSteps = 0;

	/**
	 * The contents of the agent's field of view.
	 * 
	 * The hashmap stores the information about the cells the agent sees in its perception square.
	 * The information about the individual cells is indexed according to the (x,y) coordinate used as 
	 * a key to the value in the map. The following sketch illustrates the agent's perception field:
	 * 
	 *       ***************
	 *       *      ^      *       where "s" is the sightRange number of cells.
	 *       *      s      *
	 *       *      v      *
	 *       *<-s-> a <-s->*
	 *       *      ^      *
	 *       *      s      *
	 *       *      v      *
	 *       ***************
	 */
	public HashMap<Pair<Integer, Integer>, CellType> cells = null;	

	/**
	 * Static constructor of a position pair used to access the agent's field of view.
	 * 
	 * Constructs the pair from integer positions.
	 *  
	 * @param x agent's X position in the environment
	 * @param y agent's Y position in the environmnet
	 * @return returns a new instance of Pair which can be used to access the perception hashmap
	 */
	public static Pair<Integer, Integer> pos(int x, int y) {
		return new Pair<Integer, Integer>(x,y);
	}

	/**
	 * Static constructor of a position pair used to access the agent's field of view.
	 * 
	 * Constructs the pair from string positions.
	 *  
	 * @param x agent's X position in the environment
	 * @param y agent's Y position in the environmnet
	 * @return returns a new instance of Pair which can be used to access the perception hashmap
	 */
	public static Pair<Integer, Integer> pos(String x, String y) {
		return new Pair<Integer, Integer>(Integer.parseInt(x),Integer.parseInt(y));
	}

	/**
	 * Returns the contents of the cell at the given position. 
	 */
	public CellType at(int x, int y) {
		return cells.get(pos(x,y));
	}
	
	/**
	 * Proper perception constructor.
	 * 
	 *  Constructs the instance from the XML data received from the server. For details on the implementation
	 *  check the protocol specification and correlate it with the code. 
	 */
	public Perception(Element el) {
		step=Integer.parseInt(el.getAttribute("step"));
		posX=Integer.parseInt(el.getAttribute("posx"));
		posY=Integer.parseInt(el.getAttribute("posy"));
		timestamp=Long.parseLong(el.getAttribute("id"));
		
		cells = new HashMap<Pair<Integer, Integer>, CellType>(); 

		for(Node childNode = el.getFirstChild(); childNode!=null;) {
			Node nextChild = childNode.getNextSibling();

			if (childNode.getNodeType()==Element.ELEMENT_NODE && childNode.getNodeName().equalsIgnoreCase("cell")) {
				Element cellEl = (Element) childNode;
				
				int cellX = Integer.parseInt(cellEl.getAttribute("x"));
				int cellY = Integer.parseInt(cellEl.getAttribute("y"));
				
				Node cellContentTmp = cellEl.getFirstChild();
				while (cellContentTmp.getNodeType()!=Element.ELEMENT_NODE) {
					cellContentTmp = cellContentTmp.getNextSibling();
				}
					
				Element cellContent = (Element) cellContentTmp;
				
				CellType cellType = CellType.Unknown;
				
				if (cellContent.getNodeName().equals("agent")) {
					if (cellContent.getAttribute("type").equals("enemy")) cellType = CellType.AgentEnemy; else cellType = CellType.AgentAlly;
				}
				if (cellContent.getNodeName().equals("obstacle")) cellType = CellType.Obstacle;
				if (cellContent.getNodeName().equals("cow")) cellType = CellType.Cow;
				if (cellContent.getNodeName().equals("corral")) {
					if (cellContent.getAttribute("type").equals("enemy")) cellType = CellType.CorralEnemy; else cellType = CellType.CorralAlly;
				}
				if (cellContent.getNodeName().equals("switch")) cellType = CellType.Switch;
				if (cellContent.getNodeName().equals("fence")) {
					if (cellContent.getAttribute("open").equals("true")) cellType = CellType.FenceOpen; else cellType = CellType.FenceClosed;
				}
				if (cellContent.getNodeName().equals("empty")) cellType = CellType.Empty;
				if (cellContent.getNodeName().equals("unknown")) cellType = CellType.Unknown;

				// DEBUG printout
				//System.out.println("(" + posX + "," + posY + ")" + " cellx: " + cellX + " celly: " + cellY + " type: " + cellType.toString());
				
				cells.put(pos(cellX, cellY), cellType);
			}			
			
			childNode = nextChild;
		}
	}

	/**
	 * Sets the auxiliary simulation parameters stored in the perception information.
	 * 
	 *  Extracts the information from the XML message sent to the agent at 
	 *  the beginning of the simulation.
	 */
	public static void setSimulationParams(Element el) {
		Perception.sightRange=Integer.parseInt(el.getAttribute("lineOfSight"));
		Perception.gridSizeX=Integer.parseInt(el.getAttribute("gsizex"));
		Perception.gridSizeY=Integer.parseInt(el.getAttribute("gsizey"));
		Perception.simSteps=Integer.parseInt(el.getAttribute("steps"));
		
		// DEBUG printout
		System.out.println("Simulation will take " + simSteps + " steps, the agent can see " + getSightRange() + " cells around and the size of the grid is " + gridSizeX + "x" + gridSizeY);
	}
}

